#!/bin/bash
# Fatdog Package Manager
# A puppy-compatible replacement for Puppy Package Manager
# (C) James Budiono 2012
#
# License: GNU GPL Version 3 or later
#
# 1.0 - May 2012 - features dependency checking & freespace checking
#
# Note: the bits in Global Inits, which is part of GTK-server bash init code
#       is (C) Peter van Eerten. Everything else is (C) James Budiono
#
# Note: we need bash 4 for its associative arrays
# Note: works with silent_petget, my own replacement for petget. Should also
#       work with standard petget, except that standard petget is noisy.
#
# Note to myself: gtk_tree_model_get must always be defined before use
# Note to myself: gtk_tree_store_set must always be defined before use

LANG=C # make bash run faster

### configuration
. $BOOTSTATE_PATH	# get AUFS_ROOT_ID
APPTITLE="Fatdog64 Package Manager"
PKG_DOWNLOAD_DIR=/tmp/fatdog64-packages

# pkg database file locations
MIRROR_1=http://distro.ibiblio.org/pub/linux/distributions/fatdog/pets/600/
MIRROR_2=http://ftp.nluug.nl/ibiblio/distributions/fatdog/pets/600/
MIRROR_3=http://mirror.aarnet.edu.au/pub/fatdog/pets/600/
MIRROR_4=http://ftp.cc.uoc.gr/mirrors/linux/fatdog/pets/600/
MIRROR_5=none
MIRROR_CHOICE=/root/.packages/fpm-mirror-choice
ALL_PKG_DB_SRC=Packages-puppy-fd64-offical
ALL_PKG_DB=/root/.packages/Packages-puppy-fd64-offical
INSTALLED_PKG_DB=/root/.packages/user-installed-packages
INSTALLED_PKG_DB_DIR=/root/.packages

# database fields description:
# pkgname|nameonly|version|pkgrelease|category|size|path|fullfilename|dependencies|description|
# $1-fullname $2-shortname $3-version $5-category $6-size $8-filename $9-dependencies $10-description
NOT_CATEGORISED="Not Categorised" 

### run-time variables
declare -A INSTALLED_CAT_ITERS 
declare -A ALL_CAT_ITERS 
SELECTED_MIRROR=$MIRROR_1
HIDE_CATEGORY=			# if not blank, don't use category
KEEP_PKG=				# if not blank, keep packages after install (ie don't delete it)
CHECK_DEPENDENCY=yes	# if blank, do not check dependency
DOWNLOAD_ONLY=			# if not blank, do not install packages
WORK_DIR=$(mktemp -dp /tmp fatdog-package-manager-XXXXXX)
SELECTED_ITEM_CACHE=$WORK_DIR/selected	#-all or -installed depending on which view

[ $(id -u) -ne 0 ] && exec gtksu "$APPTITLE" "$0" # must run as root

#---------------------------------------------------------- Global Inits
### global variables for proper GTK-server
declare GTK NULL="NULL"
declare PIPE=/tmp/gtk.bash.$$
trap 'gtk gtk_server_exit; exit' ERR # stop GTK-server in case of an error

### start GTK-server
gtk-server -fifo=$PIPE &
while [ ! -p $PIPE ]; do continue; done

### GTK-server helper functions
# Assignment function
function define() { $2 $3 $4 $5 $6 $7 $8 $9; eval $1="$GTK"; }

# Communicate with GTK-server
gtk()
{
/bin/echo "$@" > $PIPE
read GTK < $PIPE
}

#---------------------------------------------------------- GUI realization

function init_gtk() {
	# init gtk
	gtk gtk_init
	
	# add GTK+ API not defined in gtk-server.cfg
	gtk gtk_server_define gtk_builder_new NONE WIDGET 0
	gtk gtk_server_define gtk_builder_add_from_file NONE INT 3 WIDGET STRING NULL
	gtk gtk_server_define gtk_builder_add_from_string NONE INT 4 WIDGET STRING INT NULL
	gtk gtk_server_define gtk_builder_get_object NONE WIDGET 2 WIDGET STRING
	gtk gtk_server_define gtk_builder_connect_signals NONE NONE 2 WIDGET NULL	
	
	gtk gtk_server_define gtk_tree_store_append NONE NONE 3 WIDGET WIDGET WIDGET
	gtk gtk_server_define gtk_tree_view_expand_all NONE NONE 1 WIDGET
	gtk gtk_server_define gtk_tree_store_clear NONE NONE 1 WIDGET
	
	gtk gtk_server_define gtk_tree_view_get_cursor NONE NONE 3 WIDGET PTR_WIDGET NULL
	gtk gtk_server_define gtk_tree_path_to_string NONE STRING 1 WIDGET	
	gtk gtk_server_define gtk_tree_model_get_iter_first NONE BOOL 2 WIDGET WIDGET
	gtk gtk_server_define gtk_tree_model_iter_children NONE BOOL 3 WIDGET WIDGET WIDGET
	
	gtk gtk_server_define gtk_notebook_next_page NONE NONE 1 WIDGET	
	gtk gtk_server_define gtk_notebook_prev_page NONE NONE 1 WIDGET	
	gtk gtk_server_define gtk_label_set_markup NONE NONE 2 WIDGET STRING
	
	# hack for getting unlimited string length and leak-free operation with gtk_tree_get_model
	# see below
	gtk gtk_server_define strcat NONE STRING 2 POINTER STRING
	gtk gtk_server_define memcpy NONE NONE 3 PTR_LONG POINTER INT		
}

###################### signal handlers & helpers ########################

### helper - leak-free, unlimited string length from gtk_tree_model_get with unpatched gtk-server
# $1-tree $2-iter $3-index $4-result name
# static variable to speed up stuff
STATIC_BUF=
STATIC_ADDRESS=
function tree_model_get_string() {
	[ -z "$STATIC_BUF" ] && define STATIC_BUF gtk malloc 32 # gtk-server MAX_DIG is 32 bytes
	
	gtk gtk_server_define gtk_tree_model_get NONE NONE 5 WIDGET WIDGET INT POINTER INT # ensure correct definition
	gtk gtk_tree_model_get $1 $2 $3 $STATIC_BUF -1
	
	gtk memcpy NULL $STATIC_BUF 32	# print address stored in $BUF --> *buf
	STATIC_ADDRESS=$GTK
	if [ $STATIC_ADDRESS -ne 0 ]; then
		gtk strcat $STATIC_ADDRESS \"\" # get string point by address --> **buf	
		eval $4=\""$GTK"\"				# this will be our return value	
		gtk free $STATIC_ADDRESS		# this was dynamically alloc-ed by gtk_tree_model_get, now free it
	else
		eval $4=\"\"
	fi
}

### helper - display progress
function flush_gtk() {
	gtk gtk_events_pending
	while [ $GTK -eq 1 ]; do
		gtk gtk_main_iteration
		gtk gtk_events_pending
	done	
}
# $1-text
function display_progress() {
	gtk gtk_label_set_text $INFO_LABEL \"$1\"
	gtk gtk_widget_show $INFO_WINDOW
	flush_gtk
}
function hide_progress() {
	gtk gtk_widget_hide $INFO_WINDOW
	flush_gtk
}

### helper - download file
# $1 - filename, $2 output
function download_file() {
	display_progress "Downloading $1 ..."
	wget $SELECTED_MIRROR/$1 -O $2
}

# helper - find a package, returns an ITER_PKG pointing to that package
# $1 - pkg to find, $2-tree model returns ITER_PKG (must be freed)
function find_package() {
	local ITER ITER2 PKG
	ITER_PKG=
		
	# 1. allocate memory
	define ITER gtk gtk_server_opaque
	define ITER2 gtk gtk_server_opaque
	
	# 2. find a package that match $1 & get its dependency list 
	gtk gtk_tree_model_get_iter_first $2 $ITER
	while [ $GTK = 1 ]; do
		if [ -z "$HIDE_CATEGORY" ]; then
			gtk gtk_tree_model_iter_children $2 $ITER2 $ITER
			while [ $GTK = 1 ]; do
				tree_model_get_string $2 $ITER2 0 PKG
				if [ $PKG = $1 ]; then ITER_PKG=$ITER2; break 2; fi
				gtk gtk_tree_model_iter_next $2 $ITER2
			done
		else
			tree_model_get_string $2 $ITER 0 PKG
			if [ $PKG = $1 ]; then ITER_PKG=$ITER; break; fi
		fi
		gtk gtk_tree_model_iter_next $2 $ITER
	done
	
	# 5. done
	[ "$ITER_PKG" != "$ITER" ] && gtk free $ITER 
	[ "$ITER_PKG" != "$ITER2" ] && gtk free $ITER2
}

### helper - add categories to tree
# $1-pkg-db, $2-tree-store $3-return var-name $4-extra category item
function add_categories_to_tree() {
	local p ITER
	gtk gtk_server_define gtk_tree_store_set NONE NONE 5 WIDGET WIDGET INT STRING INT
	for p in $(awk -F"|" '{print $5}' < $1 | sort | uniq) "$4"; do
		[ -z "$p" ] && continue
		define ITER gtk gtk_server_opaque
		gtk gtk_tree_store_append $2 $ITER NULL
		gtk gtk_tree_store_set $2 $ITER 0 \""$p"\" -1
		#echo $p $ITER
		eval $3["$p"]=$ITER
	done
}

### signal handler - populate the list of all available packages
function populate_all_pkg() {
	display_progress "Refreshing list of all available packages ..."
	local ITER ITER2 OIFS 
		
	# 1. get sorted list of all categories
	[ ! -f $ALL_PKG_DB ] && download_file $ALL_PKG_DB_SRC $ALL_PKG_DB
	[ ! -f $ALL_PKG_DB ] && return
	
	# 2. add them to the tree
	gtk gtk_tree_store_clear $ALL_PKG_TREE
	[ -z "$HIDE_CATEGORY" ] && add_categories_to_tree $ALL_PKG_DB $ALL_PKG_TREE ALL_CAT_ITERS
	#declare -p ALL_CAT_ITERS

	# 3. load installed version
	OIFS=$IFS; IFS="|"; 
	declare -A installed_version
	while read p; do
		[ -z "$p" ] && continue
		set -- $p
		installed_version[$2]=$3
	done < $INSTALLED_PKG_DB
	
	# 4. add all items now
	# $1-fullname $2-shortname $3-version $5-category $6-size $8-filename $9-dependencies $10-description
	ITER=NULL
	gtk gtk_server_define gtk_tree_store_set NONE NONE 17 WIDGET WIDGET \
		INT STRING INT STRING INT STRING INT STRING INT STRING INT STRING INT STRING INT
	while read p; do
		[ -z "$p" ] && continue
		set -- $p
		[ -z "$HIDE_CATEGORY" ] && ITER=${ALL_CAT_ITERS[$5]}
		define ITER2 gtk gtk_server_opaque	
		gtk gtk_tree_store_append $ALL_PKG_TREE $ITER2 $ITER
		gtk gtk_tree_store_set $ALL_PKG_TREE $ITER2 0 \"$2\"   1 \"$6\"   2 \"${10}\"    3 \"$3\" \
													6 \"$8\"   7 \"$9\"   5 \"${installed_version[$2]}\" -1
		#echo $2 --- ${installed_version[$2]}
	done << EOF
$(cat $ALL_PKG_DB | sort -f)
EOF
	IFS=$OIFS
	gtk gtk_tree_view_expand_all $ALL_PKG_VIEW
	rm $SELECTED_ITEM_CACHE-all 2> /dev/null
	return 0 # make bash happy
}

### signal handler - populate the list of installed packages
function populate_installed_pkg() {
	display_progress "Refreshing list of installed packages ..."
	local ITER ITER2 OIFS 
		
	# 1. get sorted list of all categories
	[ ! -f $INSTALLED_PKG_DB ] && return
	
	# 2. add them to the tree
	gtk gtk_tree_store_clear $INSTALLED_PKG_TREE
	[ -z "$HIDE_CATEGORY" ] && add_categories_to_tree $INSTALLED_PKG_DB $INSTALLED_PKG_TREE INSTALLED_CAT_ITERS "$NOT_CATEGORISED"
	#declare -p INSTALLED_CAT_ITERS

	# 3. get latest versions
	OIFS=$IFS; IFS="|"; 
	declare -A latest_version
	while read p; do
		[ -z "$p" ] && continue
		set -- $p
		latest_version[$2]=$3
	done < $ALL_PKG_DB	
	
	# 4. add all items now
	# $1-fullname $2-shortname $3-version $5-category $6-size $8-filename $9-dependencies $10-description
	ITER=NULL
	gtk gtk_server_define gtk_tree_store_set NONE NONE 15 WIDGET WIDGET \
		INT STRING INT STRING INT STRING INT STRING INT STRING INT STRING INT 
	while read p; do
		[ -z "$p" ] && continue
		set -- $p
		if [ -z "$HIDE_CATEGORY" ]; then
			[ -n "$5" ] && ITER=${INSTALLED_CAT_ITERS[$5]} || ITER=${INSTALLED_CAT_ITERS[$NOT_CATEGORISED]}
		fi
		define ITER2 gtk gtk_server_opaque	
		gtk gtk_tree_store_append $INSTALLED_PKG_TREE $ITER2 $ITER
		gtk gtk_tree_store_set $INSTALLED_PKG_TREE $ITER2	0 \"$2\"  1 \"$6\"	2 \"${10}\" \
															3 \"$3\"  6 \"$8\" 	5 \"${latest_version[$2]}\" -1
		#echo $2 --- ${latest_version[$2]}
	done << EOF
$(cat $INSTALLED_PKG_DB | sort -f)
EOF
	IFS=$OIFS
	gtk gtk_tree_view_expand_all $INSTALLED_PKG_VIEW
	rm $SELECTED_ITEM_CACHE-installed 2> /dev/null
	return 0 # make bash happy
}

### signal handler - choose mirror 
function setup_mirror_dialog() {
	for idx in {1..5}; do 
		gtk gtk_button_set_label $(eval echo \$MIRROR_${idx}_RADIO) \"$(eval echo \$MIRROR_$idx)\"
	done
	
	idx=1; read idx < $MIRROR_CHOICE
	[ $idx -lt 1 -o $idx -gt 5 ] && idx=1
	gtk gtk_toggle_button_set_active $(eval echo \$MIRROR_${idx}_RADIO) 1
	SELECTED_MIRROR=$(eval echo \$MIRROR_$idx)
}
function choose_mirror() {
	local RESULT OLD_SELECTION
	OLD_SELECTION=$SELECTED_MIRROR
	define RESULT gtk gtk_dialog_run $MIRROR_DIALOG
	if [ $RESULT -eq 0 ]; then
		for idx in {1..5}; do 
			define RESULT gtk gtk_toggle_button_get_active $(eval echo \$MIRROR_${idx}_RADIO)
			if [ $RESULT -eq 1 ]; then
				SELECTED_MIRROR=$(eval echo \$MIRROR_$idx)
				echo $idx > $MIRROR_CHOICE
				break
			fi
		done
		test "$OLD_SELECTION" != "$SELECTED_MIRROR"
		return
	else
		return 1
	fi
	#echo $SELECTED_MIRROR
}

### helper - print dependencies and recurse
# $1 - package to check
# declare -A VISITED	# list of packages already check for dependencies - clear this before calling
function get_dependencies() {
	[ "${VISITED[$1]}" ] && return # don't enter again if we were already here
	local DEPENDS
	
	VISITED[$1]=1
	DEPENDS=$(awk -F"|" -v pkg=$1 '{ if ( $2 == pkg ) print $9 }' $ALL_PKG_DB)	
	DEPENDS=${DEPENDS//+/}; DEPENDS=${DEPENDS//,/ } # remove + signs, change comma to space
	if [ "$DEPENDS" ]; then
		echo $DEPENDS 
		for p in $DEPENDS; do 
			 get_dependencies $p
		done
	fi
}

# $1 item to mark, $2 value (1=on, 0=off), $3-tree, $4-type (all or install)
function mark_item() {
	local PKG PKGFILE PKGSIZE ITER_PKG
	find_package $1 $3
	if [ "$ITER_PKG" ]; then
		gtk gtk_server_define gtk_tree_store_set NONE NONE 5 WIDGET WIDGET INT INT INT 
		gtk gtk_tree_store_set $3 $ITER_PKG 4 $2 -1 	
		tree_model_get_string $3 $ITER_PKG 6 PKGFILE
		case $2 in
			1)	# turn on - add it to cache
				tree_model_get_string $3 $ITER_PKG 1 PKGSIZE
				PKGSIZE=${PKGSIZE%K*}
				echo "$PKGFILE|$PKGSIZE" >> $SELECTED_ITEM_CACHE-$4 ;;

			0)	# turn off - remove it from cache
				sed -i "/$PKGFILE/ d" $SELECTED_ITEM_CACHE-$4 ;;
		esac
		gtk free $ITER_PKG
	fi
}

### signal handler - check dependencies
function mark_dependencies() {
	display_progress "Checking dependencies ..."
	local ITER SELECTION PKG DEPENDS
	
	define ITER gtk gtk_server_opaque
	define SELECTION gtk gtk_tree_view_get_selection $ALL_PKG_VIEW
	gtk gtk_tree_selection_get_selected $SELECTION NULL $ITER
	tree_model_get_string $ALL_PKG_TREE $ITER 0 PKG
	gtk free $ITER

	unset VISITED
	declare -A VISITED
	DEPENDS=$(get_dependencies $PKG)
	DEPENDS=$(echo $DEPENDS | tr ' ' '\n' | sort -u) # sort & remove duplicates
	#echo $DEPENDS
	if [ "$DEPENDS" ]; then
		if Xdialog --title "$APPTITLE" --left --yesno "
$PKG has the following dependencies:\n\n
$(echo $DEPENDS | tr ' ' '\n')\n\n
Do you want to mark them too?
Click YES to do so, or NO if you don't want to mark them.
" 0 0;  then
			for p in $DEPENDS; do
				#echo $p
				mark_item $p 1 $ALL_PKG_TREE all
				shift
			done
		fi
	fi
	return 0 # make bash happy
}

### signal handler - check dependencies for installed packages
function mark_installed_dependencies() {
	display_progress "Checking dependencies ..."
	local ITER SELECTION PKG DEPENDS OLD_DEPENDS=""
	
	define ITER gtk gtk_server_opaque
	define SELECTION gtk gtk_tree_view_get_selection $INSTALLED_PKG_VIEW
	gtk gtk_tree_selection_get_selected $SELECTION NULL $ITER
	tree_model_get_string $INSTALLED_PKG_TREE $ITER 0 PKG
	gtk free $ITER
	
	DEPENDS=$PKG
	unset VISITED
	declare -A VISITED
	while [ "$OLD_DEPENDS" != "$DEPENDS" ]; do
		OLD_DEPENDS="$DEPENDS"
		DEPENDS=""
		for p in $OLD_DEPENDS; do
			DEPENDS="$DEPENDS $p"
			if [ -z ${VISITED[$p]} ]; then
				VISITED[$p]=1
				DEPENDS="$DEPENDS $(awk -F"|" "BEGIN { ORS=\" \"} \$9 ~ /$p/ { print \$2 }" $INSTALLED_PKG_DB)"
			fi
		done			
	done
	#echo $DEPENDS
	DEPENDS=${DEPENDS# $PKG}	# remove the original package
	DEPENDS=$(echo $DEPENDS | tr ' ' '\n' | sort -u)
	#echo $DEPENDS
	if [ "$DEPENDS" ]; then
		if Xdialog --title "$APPTITLE" --left --yesno "
$PKG is used by the following packages:\n\n
$DEPENDS\n\n
Do you want to mark them for removal too?
Click YES to do so, or NO if you don't want to mark them.
" 0 0;  then
			for p in $DEPENDS; do
				#echo $p
				mark_item $p 1 $INSTALLED_PKG_TREE installed
				shift
			done
		fi
	fi
	return 0  # make bash happy
}

### signal handler - select package
# $1-tree view, $2-tree store, $3-file-prefix, return CURRENT_ROW_SELECTED=yes or blank
function mark_selected_item() {
	local ITER RESULT SELECTION PKGFILE PKGSIZE
	
	define ITER gtk gtk_server_opaque	
	define SELECTION gtk gtk_tree_view_get_selection $1
	gtk gtk_tree_selection_get_selected $SELECTION NULL $ITER
	
	gtk gtk_server_define gtk_tree_model_get NONE NONE 5 WIDGET WIDGET INT PTR_INT INT	
	define RESULT gtk gtk_tree_model_get $2 $ITER 4 NULL -1 
	
	tree_model_get_string $2 $ITER 6 PKGFILE	
	gtk gtk_server_define gtk_tree_store_set NONE NONE 5 WIDGET WIDGET INT INT INT 
	if [ $RESULT -eq 0 ]; then
		# show to user
		gtk gtk_tree_store_set $2 $ITER 4 1 -1 
		CURRENT_ROW_SELECTED=yes
		
		tree_model_get_string $2 $ITER 1 PKGSIZE
		PKGSIZE=${PKGSIZE%K*}
		echo "$PKGFILE|$PKGSIZE" >> $SELECTED_ITEM_CACHE-$3 # store in selection cache
	else
		# show to user
		gtk gtk_tree_store_set $2 $ITER 4 0 -1 
		CURRENT_ROW_SELECTED=
		sed -i "/$PKGFILE/ d" $SELECTED_ITEM_CACHE-$3 # remove from selection cache
	fi	
	gtk free $ITER
}

### signal handler - install packages
# $* package sizes in Kb
function check_free_disk_space() {
	# get total package size
	local total=0 required=0 branch= freespace=
	while [ $1 ]; do
		#echo $1
		total=$(( $total + $1 ))
		shift
	done
	required=$(( $total * 3 )) # dotpet, extracted pet, actual install
	#echo $total $required
	
	# check available free space
	[ -z "$AUFS_ROOT_ID" ] && AUFS_ROOT_ID=$(awk '{ if ($2 == "/" && $3 == "aufs" ) print $4}' /proc/mounts | sed 's/.*si=/si_/; s/).*//') #'
	[ "$AUFS_ROOT_ID" ] && branch=$(cat /sys/fs/aufs/$AUFS_ROOT_ID/br0 | sed 's/=.*//')
	[ "$branch" ] && freespace=$(df -k | grep $branch | awk '{print $4}')
	#freespace=1
	[ $freespace -gt $total ] && return 0
	Xdialog --title "$APPTITLE" --infobox "
Not enough space to install packages.
Your total package size is $total KB, which requires $required KB,
but you only have $freespace KB in your savefile.

Please adjust your selection or uninstall some packages first.
" 0 0 10000
	return 1
}
function install_packages() {
	local SELECTED_PACKAGES=$(cat $SELECTED_ITEM_CACHE-all 2> /dev/null | awk -F"|" '{print $1}')
	! [ "$SELECTED_PACKAGES" ] && return 0
	if Xdialog --title "$APPTITLE" --left --yesno "
Install the following packages?\n\n
$(echo $SELECTED_PACKAGES | tr ' ' '\n')\n\n
Click YES to continue to install, NO to go back and modify your selection.
" 0 0; then
		check_free_disk_space $(cat $SELECTED_ITEM_CACHE-all 2> /dev/null | awk -F"|" '{print $2}') || return 0
		for p in $SELECTED_PACKAGES; do
			download_file $p "$PKG_DOWNLOAD_DIR/$p"
			if [ -z $DOWNLOAD_ONLY ]; then
				display_progress "Installing $p ..."
				echo install $p
				silent_petget "$PKG_DOWNLOAD_DIR/$p"
			fi
			[ -z "$KEEP_PKG" ] && rm "$PKG_DOWNLOAD_DIR/$p"
		done
		fixmenus
		
		# done, refresh list and tell user
		populate_installed_pkg
		populate_all_pkg
		hide_progress
		if [ "$KEEP_PKG" ]; then
			su $USER -c "rox \"$PKG_DOWNLOAD_DIR\""
			Xdialog --title "$APPTITLE" --infobox "Installation complete.
Your packages are downloaded in $PKG_DOWNLOAD_DIR, which has just been opened.
Note that this is a temporary location, move them elsewhere if you want to keep them permanently.
	" 0 0 10000
		else
			Xdialog --title "$APPTITLE" --infobox "Installation complete.." 0 0 10000
		fi
	fi
	return 0 # make bash happy
}

### signal handler - remove packages
function remove_packages() {
	local SELECTED_PACKAGES=$(cat $SELECTED_ITEM_CACHE-installed 2> /dev/null | awk -F"|" '{print $1}')
	! [ "$SELECTED_PACKAGES" ] && return 0
	if Xdialog --title "$APPTITLE" --left --yesno "
Remove the following packages?\n\n
$(echo $SELECTED_PACKAGES | tr ' ' '\n')\n\n
Click YES to continue to remove, NO to go back and modify your selection.
" 0 0; then
		for p in $SELECTED_PACKAGES; do
			display_progress "Un-installing $p ..."
			echo uninstall $p 
			silent_petget $p uninstall 
		done
		fixmenus
		
		# done, refresh list and tell user
		populate_installed_pkg
		populate_all_pkg
		hide_progress
		Xdialog --title "$APPTITLE" --infobox "Un-installation complete.." 0 0 10000
	fi
	return 0 # make sure bash is happy	
}

### signal handler - showfiles in highlighted package
function showfiles_installed() {
	local SELECTED_PATH	ITER PKGFILE
	define ITER gtk gtk_server_opaque
	
	PKGFILE=""
	define SELECTED_PATH gtk gtk_tree_view_get_cursor $INSTALLED_PKG_VIEW NULL
	if [ $SELECTED_PATH -ne 0 ]; then
		gtk gtk_tree_model_get_iter $INSTALLED_PKG_TREE $ITER $SELECTED_PATH	
		gtk gtk_tree_path_free $SELECTED_PATH

		# $1-tree $2-iter $3-index $4-result name
		tree_model_get_string $INSTALLED_PKG_TREE $ITER 6 PKGFILE
	fi
	gtk free $ITER
	echo $PKGFILE
	if [ "$PKGFILE" ]; then
		PKGFILE=${PKGFILE%.pet}.files
		Xdialog --title "$APPTITLE" --backtitle "Files installed by ${PKGFILE%.files}" \
		        --no-cancel --textbox "$INSTALLED_PKG_DB_DIR/$PKGFILE" 0 0 
	fi
	return 0 # make sure bash is happy	
}

function cleanup() {
	rm -rf $WORK_DIR
}

#############################   main   ##############################
### load GUI
init_gtk
define BUILDER gtk gtk_builder_new
define RESULT gtk gtk_builder_add_from_file $BUILDER "${0}.xml"
gtk gtk_builder_connect_signals $BUILDER

define MAIN_WINDOW gtk gtk_builder_get_object $BUILDER main_window
define HIDE_CATEGORY_CHECKBOX gtk gtk_builder_get_object $BUILDER hide-category
define KEEP_PKG_CHECKBOX gtk gtk_builder_get_object $BUILDER keep-pkg
define CHECK_DEPENDENCY_CHECKBOX gtk gtk_builder_get_object $BUILDER check-dependency
define DOWNLOAD_ONLY_CHECKBOX gtk gtk_builder_get_object $BUILDER download-only

define ALL_PKG_TREE gtk gtk_builder_get_object $BUILDER allpackagestree
define ALL_PKG_VIEW gtk gtk_builder_get_object $BUILDER allpackagesview
define ALL_PKG_REFRESH_BUTTON gtk gtk_builder_get_object $BUILDER refresh-all
define CHOOSE_MIRROR_BUTTON gtk gtk_builder_get_object $BUILDER choose-mirror
define INSTALL_BUTTON gtk gtk_builder_get_object $BUILDER install
define QUIT_ALL_BUTTON gtk gtk_builder_get_object $BUILDER quit-all

define INSTALLED_PKG_TREE gtk gtk_builder_get_object $BUILDER installedpackagestree
define INSTALLED_PKG_VIEW gtk gtk_builder_get_object $BUILDER installedpackagesview
define INSTALLED_PKG_REFRESH_BUTTON gtk gtk_builder_get_object $BUILDER refresh-installed
define UNINSTALL_BUTTON gtk gtk_builder_get_object $BUILDER uninstall
define SHOWFILES_BUTTON gtk gtk_builder_get_object $BUILDER showfiles
define QUIT_INSTALLED_BUTTON gtk gtk_builder_get_object $BUILDER quit-installed

define MIRROR_DIALOG gtk gtk_builder_get_object $BUILDER mirror_dialog
define MIRROR_1_RADIO gtk gtk_builder_get_object $BUILDER mirror-1
define MIRROR_2_RADIO gtk gtk_builder_get_object $BUILDER mirror-2
define MIRROR_3_RADIO gtk gtk_builder_get_object $BUILDER mirror-3
define MIRROR_4_RADIO gtk gtk_builder_get_object $BUILDER mirror-4
define MIRROR_5_RADIO gtk gtk_builder_get_object $BUILDER mirror-5

define INFO_WINDOW gtk gtk_builder_get_object $BUILDER info_window
define INFO_LABEL gtk gtk_builder_get_object $BUILDER info-label

gtk g_object_unref $BUILDER

### connect signals
gtk gtk_server_connect $MAIN_WINDOW delete-event finish
gtk gtk_server_connect $HIDE_CATEGORY_CHECKBOX clicked hide-category-clicked
gtk gtk_server_connect $KEEP_PKG_CHECKBOX clicked keep-pkg-clicked
gtk gtk_server_connect $CHECK_DEPENDENCY_CHECKBOX clicked check-dependency-clicked
gtk gtk_server_connect $DOWNLOAD_ONLY_CHECKBOX clicked download-only-clicked

gtk gtk_server_connect $ALL_PKG_REFRESH_BUTTON clicked refresh-all-clicked
gtk gtk_server_connect $CHOOSE_MIRROR_BUTTON clicked choose-mirror-clicked
gtk gtk_server_connect $INSTALL_BUTTON clicked install-clicked
gtk gtk_server_connect $QUIT_ALL_BUTTON clicked quit-clicked
gtk gtk_server_connect $ALL_PKG_VIEW row-activated all-select-pkg

gtk gtk_server_connect $INSTALLED_PKG_REFRESH_BUTTON clicked refresh-installed-clicked
gtk gtk_server_connect $UNINSTALL_BUTTON clicked uninstall-clicked
gtk gtk_server_connect $SHOWFILES_BUTTON clicked showfiles-clicked
gtk gtk_server_connect $QUIT_INSTALLED_BUTTON clicked quit-clicked
gtk gtk_server_connect $INSTALLED_PKG_VIEW row-activated installed-select-pkg

### initialisation stuff
mkdir -p $PKG_DOWNLOAD_DIR
chmod a+rwx $PKG_DOWNLOAD_DIR
setup_mirror_dialog
populate_all_pkg
populate_installed_pkg

### show main window and enter main loop
gtk gtk_widget_show $MAIN_WINDOW
while true; do 
	hide_progress
	define EVENT gtk gtk_server_callback wait
	case $EVENT in
		finish|quit-clicked) break ;;
		refresh-all-clicked)
			rm $ALL_PKG_DB	# force re-download
			populate_all_pkg ;;
			
		refresh-installed-clicked)	
			populate_installed_pkg ;;

		showfiles-clicked)
			showfiles_installed ;;
			
		hide-category-clicked)
			HIDE_CATEGORY=
			gtk gtk_toggle_button_get_active $HIDE_CATEGORY_CHECKBOX
			[ $GTK -ne 0 ] && HIDE_CATEGORY=yes 
			populate_all_pkg
			populate_installed_pkg ;;
			
		keep-pkg-clicked)
			KEEP_PKG=yes
			gtk gtk_toggle_button_get_active $KEEP_PKG_CHECKBOX
			if [ $GTK -eq 0 ]; then # not active
				KEEP_PKG=
				DOWNLOAD_ONLY=
				gtk gtk_toggle_button_set_active $DOWNLOAD_ONLY_CHECKBOX 0
			fi ;;
			
		download-only-clicked)
			DOWNLOAD_ONLY=
			gtk gtk_toggle_button_get_active $DOWNLOAD_ONLY_CHECKBOX
			if [ $GTK -ne 0 ]; then # active
				DOWNLOAD_ONLY=yes
				KEEP_PKG=yes				
				gtk gtk_toggle_button_set_active $KEEP_PKG_CHECKBOX 1
			fi ;;
			
		check-dependency-clicked)
			CHECK_DEPENDENCY=
			gtk gtk_toggle_button_get_active $CHECK_DEPENDENCY_CHECKBOX
			[ $GTK -ne 0 ] && CHECK_DEPENDENCY=yes ;;		
						
		choose-mirror-clicked)
			if choose_mirror; then
				rm $ALL_PKG_DB	# force re-download
				populate_all_pkg
				populate_installed_pkg 
			fi ;;
			
		all-select-pkg)
			mark_selected_item $ALL_PKG_VIEW $ALL_PKG_TREE all
			[ "$CHECK_DEPENDENCY" -a "$CURRENT_ROW_SELECTED" ] && mark_dependencies ;;
		installed-select-pkg)		
			mark_selected_item $INSTALLED_PKG_VIEW $INSTALLED_PKG_TREE installed 
			[ "$CHECK_DEPENDENCY" -a "$CURRENT_ROW_SELECTED" ] && mark_installed_dependencies ;;
		
		install-clicked)
			install_packages ;;
		uninstall-clicked)
			remove_packages ;;
	esac
done
gtk gtk_server_exit 
cleanup
